"""
    Base utilities related to LXD and its settings
    ==============================================
    This module provides basic tools allowing to retrieve LXD settings / configuration options or to
    interract with LXD...
"""

import os


def get_lxd_dir():
    """ Returns the path (as a string) towards the LXD's directory. """
    return os.environ.get('LXD_DIR', None) or '/var/lib/lxd'



"""
    Host components
    ===============
    This package contains host definitions. Hosts are used to perform operations on the "host" side
    and can be used to interact with the host platform.
"""

from .base import *  # noqa



from .base import Guest


class OpenSUSEGuest(Guest):
    """ This guest can provision openSUSE containers. """

    name = 'opensuse'
    barebones_packages = [
        'openSSH',
        'python3-base',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['zypper', '--non-interactive', 'install', ] + self.barebones_packages)



from .base import Guest


class ArchLinuxGuest(Guest):
    """ This guest can provision ArchLinux containers. """

    name = 'arch'
    barebones_packages = [
        'openssh',
        'python2',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['pacman', '-S', '--noconfirm'] + self.barebones_packages)



import logging
import os

import yaml
from voluptuous.error import Invalid

from . import constants
from .exceptions import ConfigFileNotFoundError, ConfigFileValidationError
from .schema import schema

logger = logging.getLogger(__name__)


class Config(object):
    """ Holds the configuration of a LXDock project. """

    def __init__(self, base_dir, filename):
        self.base_dir = base_dir
        self.filename = filename
        self.containers = []
        self._dict = {}

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        return self._dict[key]

    @classmethod
    def from_base_dir(cls, base_dir='.'):
        """ Returns a Config instance using a base directory. """
        existing_config_filenames = [
            filename for filename in constants.ALLOWED_FILENAMES
            if os.path.exists(os.path.join(base_dir, filename))]

        if not existing_config_filenames:
            raise ConfigFileNotFoundError(
                'Unable to find a suitable configuration file in this directory. '
                'Are you in the right directory?\n'
                'The supported filenames are: {}'.format(', '.join(constants.ALLOWED_FILENAMES))
            )

        config_filename = os.path.join(base_dir, existing_config_filenames[0])
        if len(existing_config_filenames) > 1:
            logger.warn('Multiple config files were found: {0}'.format(
                ', '.join(existing_config_filenames)))
            logger.warn('Using: {0}'.format(config_filename))

        # Initializes the config instance.
        config = cls(base_dir, config_filename)

        # Loads the YML.
        config.load()

        # Validates the content of the configuration.
        try:
            schema(config._dict)
        except Invalid as e:
            # Formats the voluptuous error
            path = ' @ %s' % '.'.join(map(str, e.path)) if e.path else ''
            msg = 'The LXDock file is invalid because: {0}'.format(e.msg + path)
            raise ConfigFileValidationError(msg)

        # Loads the containers.
        config.load_containers()

        return config

    def load(self):
        """ Loads the YML configuration file and store it inside the `_dict` dictionary. """
        self._dict = self._load_yml()

    def load_containers(self):
        """ Loads each container configuration and store it inside the `containers` attribute. """
        containers = [ContainerConfig(self._get_container_config_dict(cdict))
                      for cdict in self._dict.get('containers', [])]
        # If we cannot consider multiple containers, we just pass the full dictionary to initialize
        # the `ContainerConfig` instance.
        if not len(containers):
            unique_container_config = ContainerConfig(self._dict)
            # We associate a specific name to the unique container if it wasn't explicitly defined
            # under the 'containers' section of the LXDock file.
            unique_container_config['name'] = 'default'
            containers = [unique_container_config, ]

        self.containers.extend(containers)

    def serialize(self):
        """ Returns the configuration as a string. """
        return yaml.dump(self._dict, default_flow_style=False)

    def _get_container_config_dict(self, container_dict):
        """ Returns a dictionary containing the container's configuration.

        The dictionary that is returned contains the container's configuration and the global config
        values. These global values can be defined outside of the scope of the container's config
        and can be used by each container.
        """
        container_config = dict(self._dict)
        container_config.update(container_dict)
        del container_config['containers']
        return container_config

    def _load_yml(self):
        """ Loads the YML configuration file. """
        with open(self.filename, 'r') as fp:
            return yaml.safe_load(fp)
        # TODO: handle potential errors here


class ContainerConfig(dict):
    """ Holds the specific configuration of a container. """



from .base import Guest


class OracleLinuxGuest(Guest):
    """ This guest can provision Oracle Linux containers. """

    name = 'ol'
    barebones_packages = [
        'openssh-server',
        'python',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['yum', '-y', 'install'] + self.barebones_packages)



import unittest.mock

import pytest
from pylxd.exceptions import NotFound

from lxdock.guests import Guest
from lxdock.guests.base import InvalidGuest


class TestGuest:
    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_guest_name(self):
        with pytest.raises(InvalidGuest):
            class InvalidDummyGuest(Guest):
                name = None

    def test_can_detect_os_or_distribution(self):
        class DummyGuest(Guest):
            name = 'dummy'
        lxd_container_1 = unittest.mock.Mock()
        lxd_container_1.files.get.return_value = 'ID=dummy'
        lxd_container_2 = unittest.mock.Mock()
        lxd_container_2.files.get.return_value = 'ID=unknown'
        lxd_container_3 = unittest.mock.Mock()
        lxd_container_3.files.get.side_effect = NotFound(response=unittest.mock.Mock())
        assert DummyGuest.detect(lxd_container_1)
        assert not DummyGuest.detect(lxd_container_2)
        assert not DummyGuest.detect(lxd_container_3)

    def test_can_add_ssh_pubkey_to_root_authorized_keys(self):
        class DummyGuest(Guest):
            name = 'dummy'
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = DummyGuest(lxd_container)
        guest.add_ssh_pubkey_to_root_authorized_keys('pubkey')
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == (['mkdir', '-p', '/root/.ssh'], )
        assert lxd_container.files.put.call_count == 1
        assert lxd_container.files.put.call_args[0] == ('/root/.ssh/authorized_keys', 'pubkey', )

    def test_can_create_a_user_with_a_default_home_directory(self):
        class DummyGuest(Guest):
            name = 'dummy'
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = DummyGuest(lxd_container)
        guest.create_user('usertest')
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == (['useradd', '--create-home', 'usertest'], )

    def test_can_create_a_user_with_a_custom_home_directory(self):
        class DummyGuest(Guest):
            name = 'dummy'
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = DummyGuest(lxd_container)
        guest.create_user('usertest', '/opt/usertest')
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == \
            (['useradd', '--create-home', '--home-dir', '/opt/usertest', 'usertest'], )



import pytest

from lxdock.provisioners import Provisioner
from lxdock.provisioners.base import InvalidProvisioner


class TestProvisioner:
    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_provisioner_name(self):
        with pytest.raises(InvalidProvisioner):
            class InvalidDummyHost(Provisioner):
                name = None



import io
import re
import subprocess
import tempfile


def get_ipv4_ip(container):
    """ Returns the IP adress of a specific container. """
    state = container.state()
    if state.network is None:  # container is not running
        return ''
    eth0 = state.network['eth0']
    for addr in eth0['addresses']:
        if addr['family'] == 'inet':
            return addr['address']
    return ''


RE_ETCHOST_LINE = re.compile(r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+([\w\-_.]+)$')


class EtcHostsBase:
    def __init__(self, etchosts_fp):
        self.lines = etchosts_fp.readlines()
        self.changed = False
        self.lxdock_bindings = {}
        self.lxdock_section_begin = None
        self.lxdock_section_end = None
        for i, line in enumerate(self.lines):
            if self.lxdock_section_begin is None:
                if line.startswith('# BEGIN LXDock section'):
                    self.lxdock_section_begin = i
            elif self.lxdock_section_end is None:
                if line.startswith('# END LXDock section'):
                    self.lxdock_section_end = i
                else:
                    m = RE_ETCHOST_LINE.match(line.strip())
                    if m:
                        self.lxdock_bindings[m.group(2)] = m.group(1)
            else:
                break

    def ensure_binding_present(self, hostname, target_ip):
        if self.lxdock_bindings.get(hostname) != target_ip:
            self.lxdock_bindings[hostname] = target_ip
            self.changed = True

    def ensure_binding_absent(self, hostname):
        if hostname in self.lxdock_bindings:
            del self.lxdock_bindings[hostname]
            self.changed = True

    def get_mangled_contents(self):
        tosave = self.lines[:]
        if self.lxdock_bindings:
            toinsert = ['# BEGIN LXDock section\n']
            toinsert += ['{} {}\n'.format(ip, host) for host, ip in self.lxdock_bindings.items()]
            toinsert.append('# END LXDock section\n')
        else:
            toinsert = []
        if self.lxdock_section_begin is not None:
            # Replace the current lxdock section with our new hosts
            begin = self.lxdock_section_begin
            end = self.lxdock_section_end + 1 if self.lxdock_section_end is not None else None
            tosave[begin:end] = toinsert
        else:
            # Append a new lxdock section at the end of the file
            tosave += toinsert
        return tosave


class EtcHosts(EtcHostsBase):
    def __init__(self, path='/etc/hosts'):
        self.path = path
        etchosts_fp = open(self.path, 'rt', encoding='utf-8')
        super().__init__(etchosts_fp)

    def save(self):
        tosave = self.get_mangled_contents()
        # First, let's try to write the file directly. Who knows, it might work!
        try:
            with open(self.path, 'wt', encoding='utf-8') as fp:
                fp.writelines(tosave)
        except PermissionError:
            # Ok, we don't have permission to /etc/hosts. Let's save it to a temp file and
            # "sudo cp" it to /etc/hosts
            with tempfile.NamedTemporaryFile('wt', encoding='utf-8') as fp:
                fp.writelines(tosave)
                fp.flush()
                cmd = "sudo cp {} {}".format(fp.name, self.path)
                p = subprocess.Popen(cmd, shell=True)
                p.wait()


class ContainerEtcHosts(EtcHostsBase):
    def __init__(self, container, path='/etc/hosts'):
        self.path = path
        self.container = container
        etchosts_contents = container.files.get(path).decode('utf-8')
        etchosts_fp = io.StringIO(etchosts_contents)
        super().__init__(etchosts_fp)

    def save(self):
        tosave = self.get_mangled_contents()
        towrite = ''.join(tosave).encode('utf-8')
        self.container.files.put(self.path, towrite)



import unittest.mock

from lxdock.cli.utils import yesno


class TestYesNoFunction(object):
    def test_can_accept_yes_answers(self):
        with unittest.mock.patch('builtins.input', return_value='Y'):
            assert yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='y'):
            assert yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='Yes'):
            assert yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='yes'):
            assert yesno('True or False?')

    def test_can_accept_no_answers(self):
        with unittest.mock.patch('builtins.input', return_value='N'):
            assert not yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='n'):
            assert not yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='No'):
            assert not yesno('True or False?')
        with unittest.mock.patch('builtins.input', return_value='no'):
            assert not yesno('True or False?')



import unittest.mock

from lxdock.guests import GentooGuest


class TestGentooGuest:
    def test_should_install_barebones_packages_if_not_installed(self):
        lxd_container = unittest.mock.Mock()
        # Retcode 1: "No installed packages matching {keyword}"
        lxd_container.execute.return_value = (1, 'ok', '')
        guest = GentooGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 1 + 2 * len(GentooGuest.barebones_packages)
        assert lxd_container.execute.call_args_list[0][0] == \
            (['emerge', 'app-portage/gentoolkit'], )
        for i, p in enumerate(GentooGuest.barebones_packages):
            assert lxd_container.execute.call_args_list[2 * i + 1][0] == (['equery', 'list', p], )
            assert lxd_container.execute.call_args_list[2 * i + 2][0] == (['emerge', p], )

    def test_should_not_reinstall_barebones_packages_if_already_installed(self):
        lxd_container = unittest.mock.Mock()
        # Retcode 0: the package has been found locally
        lxd_container.execute.return_value = (0, 'ok', '')
        guest = GentooGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 1 + len(GentooGuest.barebones_packages)
        assert lxd_container.execute.call_args_list[0][0] == \
            (['emerge', 'app-portage/gentoolkit'], )
        for i, p in enumerate(GentooGuest.barebones_packages):
            assert lxd_container.execute.call_args_list[i + 1][0] == (['equery', 'list', p], )



import unittest.mock

from lxdock.guests import AlpineGuest


class TestAlpineGuest:
    def test_can_install_barebones_packages(self):
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = AlpineGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 4
        assert lxd_container.execute.call_args_list[0][0] == (['apk', 'update'], )
        assert lxd_container.execute.call_args_list[1][0] == \
            (['apk', 'add'] + AlpineGuest.barebones_packages, )
        assert lxd_container.execute.call_args_list[2][0] == \
            (['rc-update', 'add', 'sshd'], )
        assert lxd_container.execute.call_args_list[3][0] == \
            (['/etc/init.d/sshd', 'start'], )



import unittest.mock

from lxdock.guests import FedoraGuest


class TestFedoraGuest:
    def test_can_install_barebones_packages(self):
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = FedoraGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == \
            (['dnf', '-y', 'install'] + FedoraGuest.barebones_packages, )



"""
    Guest components
    ================
    This package contains guest definitions. Guests are used to perform operations on the "guest"
    side (the containers).
"""

from .alpine import *     # noqa
from .archlinux import *  # noqa
from .base import *       # noqa
from .centos import *     # noqa
from .debian import *     # noqa
from .fedora import *     # noqa
from .gentoo import *     # noqa
from .opensuse import *   # noqa
from .oracle import *     # noqa
from .ubuntu import *     # noqa



import re

from voluptuous.schema_builder import message
from voluptuous.validators import truth


hostname_part_re = re.compile(r'(?!-)[A-Z\d-]{1,63}(?<!-)$', re.IGNORECASE)
lxd_identifier_re = re.compile(r'^(([A-Z]|[A-Z][A-Z0-9\-]*[A-Z0-9])\.)*([A-Z]|[A-Z][A-Z0-9\-]*[A-Z0-9])$', re.IGNORECASE)  # noqa


@message('expected a valid hostname')
@truth
def Hostname(v):
    """ Validates a hostname. """
    if len(v) > 255:
        return False
    # Strips one dot from the right if applicable.
    v = v[:-1] if v.endswith('.') else v
    # Validates each part of the hostname.
    return all(hostname_part_re.match(part) for part in v.split('.'))


@message(
    'expected a valid identifer no longer than 63 characters, starting with letters and made up of '
    'letters, digits and dashes')
@truth
def LXDIdentifier(v):
    """ Validates an identifier that should obey to the same rules as RFC 952 hostnames. """
    if len(v) > 63:
        return False
    return lxd_identifier_re.match(v)



import os

from ..client import get_client
from ..conf import Config
from ..project import Project


def get_project(base_dir='.'):
    """ Returns a fully functional `Project` instance that will be used by the CLI tools. """
    # Fetches the config options from the current directory.
    # Note/TODO: this could be improved by adding a parameter allowing to manually set the base dir.
    config = Config.from_base_dir(base_dir)

    # Determines the name of the project from the config file.
    project_name = config['name']

    # Initializes the client instance that will be used to orchestrate containers.
    client = get_client()

    return Project.from_config(project_name, os.path.abspath(base_dir), client, config)



import os
import types
import unittest.mock

from lxdock import constants
from lxdock.container import Container, must_be_running
from lxdock.test.testcases import LXDTestCase


THIS_DIR = os.path.join(os.path.dirname(__file__))


def test_must_be_running_decorator_works(persistent_container):
    @must_be_running
    def dummy_action(self):
        return 42
    persistent_container.dummy_action = types.MethodType(dummy_action, persistent_container)
    persistent_container.halt()
    assert persistent_container.dummy_action() is None
    persistent_container.up()
    assert persistent_container.dummy_action() == 42
    del persistent_container.dummy_action


class TestContainer(LXDTestCase):
    def test_can_set_up_a_container_that_does_not_exist(self):
        container_options = {
            'name': self.containername('dummy'), 'image': 'ubuntu/xenial', 'mode': 'pull', }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        assert container._container.status_code == constants.CONTAINER_RUNNING
        assert container._container.config['user.lxdock.made'] == '1'
        assert container._container.config['user.lxdock.homedir'] == THIS_DIR

    def test_can_set_up_a_container_that_is_already_up_and_running(self, persistent_container):
        persistent_container.up()
        assert persistent_container._container.status_code == constants.CONTAINER_RUNNING

    def test_can_set_up_a_container_that_exists_but_is_not_running(self, persistent_container):
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED
        persistent_container.up()
        assert persistent_container._container.status_code == constants.CONTAINER_RUNNING

    def test_can_destroy_a_container_and_run_this_action_for_a_container_that_does_not_exist(self):
        container_options = {
            'name': self.containername('dummy'), 'image': 'ubuntu/xenial', 'mode': 'pull', }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.destroy()
        assert not container.exists
        container.up()
        assert container.exists
        container.destroy()
        assert not container.exists

    def test_can_halt_a_container_that_is_running(self, persistent_container):
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED

    def test_can_try_to_halt_a_container_that_is_already_stoppeds(self, persistent_container):
        persistent_container.halt()
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED

    def test_can_provision_a_container(self):
        container_options = {
            'name': self.containername('dummy'), 'image': 'ubuntu/xenial', 'mode': 'pull',
            'provisioning': [
                {'type': 'ansible',
                 'playbook': os.path.join(THIS_DIR, 'fixtures/provision_with_ansible.yml'), }
            ],
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        assert container._container.config['user.lxdock.provisioned'] == 'true'
        assert container._container.files.get('/dummytest').strip() == b'dummytest'

    @unittest.mock.patch('subprocess.call')
    def test_can_open_a_shell_for_the_root_user(self, mocked_call, persistent_container):
        persistent_container.shell()
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} -- su -m root'.format(persistent_container.lxd_name)

    @unittest.mock.patch('subprocess.call')
    def test_can_open_a_shell_for_a_specific_shelluser(self, mocked_call):
        container_options = {
            'name': self.containername('dummy'), 'image': 'ubuntu/xenial', 'mode': 'pull',
            'shell': {'user': 'test', 'home': '/opt', },
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        container.shell()
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} --env HOME=/opt -- su -m test'.format(container.lxd_name)

    def test_can_tell_if_a_container_exists_or_not(self, persistent_container):
        unkonwn_container = Container('myproject', THIS_DIR, self.client, **{
            'name': self.containername('unkonwn'), 'image': 'ubuntu/xenial', 'mode': 'pull', })
        assert persistent_container.exists
        assert not unkonwn_container.exists

    def test_can_tell_if_a_container_is_privileged_or_not(self, persistent_container):
        persistent_container._container.config['security.privileged'] = 'true'
        persistent_container._container.save(wait=True)
        assert persistent_container.is_privileged
        persistent_container._container.config['security.privileged'] = 'false'
        persistent_container._container.save(wait=True)
        assert not persistent_container.is_privileged

    def test_can_tell_if_a_container_is_provisioned_or_not(self, persistent_container):
        persistent_container._container.config['user.lxdock.provisioned'] = 'false'
        persistent_container._container.save(wait=True)
        assert not persistent_container.is_provisioned
        persistent_container.provision()
        assert persistent_container.is_provisioned

    def test_can_tell_if_a_container_is_running_or_not(self, persistent_container):
        persistent_container.halt()
        assert not persistent_container.is_running
        persistent_container.up()
        assert persistent_container.is_running

    def test_can_tell_if_a_container_is_stopped_or_not(self, persistent_container):
        persistent_container.halt()
        assert persistent_container.is_stopped
        persistent_container.up()
        assert not persistent_container.is_stopped

    def test_can_return_its_status(self, persistent_container):
        unkonwn_container = Container('myproject', THIS_DIR, self.client, **{
            'name': self.containername('unkonwn'), 'image': 'ubuntu/xenial', 'mode': 'pull', })
        assert unkonwn_container.status == 'not-created'
        persistent_container.halt()
        assert persistent_container.status == 'stopped'
        persistent_container.up()
        assert persistent_container.status == 'running'



from .base import Guest


class DebianGuest(Guest):
    """ This guest can provision Debian containers. """

    name = 'debian'
    barebones_packages = [
        'apt-utils',
        'aptitude',
        'openssh-server',
        'python',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['apt-get', 'install', '-y'] + self.barebones_packages)



"""
    Base host
    =========
    This module provides the `Host` base class that is used to define the host OS or distributions
    supported by LXDock.
"""

import os
import platform
import subprocess
from pathlib import Path

from ..utils.lxd import get_lxd_dir
from ..utils.metaclass import with_metaclass

__all__ = ['Host', ]


class InvalidHost(Exception):
    """ The `Host` subclass is not valid. """


class _HostBase(type):
    """ Metaclass for all LXD hosts.

    This metaclass ensures that all defined `Host` subclasses have the required attributes and
    proceeds to some validation checks. Additionally it implements the "plugin mount" paradigm and
    stores a list of `Host` subclasses in the namespace of the "plugin mount" class (`host`).
    """

    def __new__(cls, name, bases, attrs):
        super_new = super(_HostBase, cls).__new__
        parents = [base for base in bases if isinstance(base, _HostBase)]

        # We stop here if we are considering the top-level class to which this metaclass was applied
        # and not one of its subclasses (eg. Host).
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Constructs the Host class.
        new_host = super_new(cls, name, bases, attrs)

        # Performs some validation checks.
        if not new_host.name:
            raise InvalidHost("The 'name' attribute of Host subclasses cannot be None")

        return new_host

    def __init__(cls, name, bases, attrs):
        # We implement the "mount point" paradigm here in order to make all `Host` subclassses
        # available in a single attribute.
        if not hasattr(cls, 'hosts'):
            # The class has no hosts list: this means that we are considering the "plugin mount"
            # class. So we created the list that will hold all the defined `Host` subclasses.
            cls.hosts = []
        else:
            # The `hosts` attribute already exists so we are considering a `Host` subclass, which
            # needs to be registered.
            cls.hosts.append(cls)


class Host(with_metaclass(_HostBase)):
    """ Represents a single host.

    `Host` subclasses will be used by `Container` instances to perform common operations on the
    host side. For example they can be used to retrieve some date (SSH pukeys, ...) or to set up
    contaianers' hosts in the /etc/hosts file. `Host` subclasses should correspond to specific OSes
    or distributions that can be used to run LXD and LXDock.
    """

    # The `name` of a host is a required attribute and should always be set on `Host` subclasses.
    name = None

    def __init__(self, lxd_container):
        self.lxd_container = lxd_container

    @classmethod
    def detect(cls):
        """ Detects if the host is an "instance" of the considered OS/distribution. """
        return cls.name.lower() in platform.platform().lower()

    def get_ssh_pubkey(self):
        """ Returns the SSH public key of the current user or None if it cannot be found. """
        pubkey_path = Path(os.path.expanduser('~/.ssh/id_rsa.pub'))
        try:
            return pubkey_path.open().read()
        except FileNotFoundError:  # pragma: no cover
            pass

    def give_current_user_access_to_share(self, source):
        """ Give read/write access to `source` for the current user. """
        subprocess.Popen('setfacl -Rdm u:{}:rwX {}'.format(os.getuid(), source), shell=True).wait()

    def give_mapped_user_access_to_share(self, source, userpath=None):
        """ Give read/write access to `source` for the mapped user owning `userpath`.

        `userpath` is a path that is relative to the LXD base directory (where LXD store contaners).
        """
        # LXD uses user namespaces when running safe containers. This means that it maps a set of
        # uids and gids on the host to a set of uids and gids in the container.
        # When considering unprivileged containers we want to ensure that the "root user" (or any
        # other user) of such containers have the proper rights to write in shared folders. To do so
        # we have to retrieve the UserID on the host-side that is mapped to the "root"'s UserID (or
        # any other user's UserID) on the guest-side. This will allow to set ACL on the host-side
        # for this UID. By doing this we will also allow "root" user on the guest-side to read/write
        # in shared folders.
        container_path_parts = [get_lxd_dir(), 'containers', self.lxd_container.name, 'rootfs']
        container_path_parts += userpath.split('/') if userpath else []
        container_path = os.path.join(*container_path_parts)
        container_path_stats = os.stat(container_path)
        host_userpath_uid = container_path_stats.st_uid
        subprocess.Popen(
            'setfacl -Rm user:lxd:rwx,default:user:lxd:rwx,'
            'user:{0}:rwx,default:user:{0}:rwx {1}'.format(host_userpath_uid, source),
            shell=True).wait()



import logging
import os
import subprocess
import time
from functools import wraps

from pylxd.exceptions import LXDAPIException, NotFound

from . import constants
from .exceptions import ContainerOperationFailed
from .guests import Guest
from .hosts import Host
from .network import EtcHosts, get_ipv4_ip
from .provisioners import Provisioner
from .utils.identifier import folderid

logger = logging.getLogger(__name__)


def must_be_running(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.is_running:
            logger.error('The container is not running.')
            return
        return method(self, *args, **kwargs)

    return wrapper


class Container:
    """ Represents a specific container that is managed by LXDock. """

    # The default image server that will be used to pull images in "pull" mode.
    _default_image_server = 'https://images.linuxcontainers.org'

    def __init__(self, project_name, homedir, client, **options):
        self.project_name = project_name
        self.homedir = homedir
        self.client = client
        self.options = options

    #####################
    # CONTAINER ACTIONS #
    #####################

    def destroy(self):
        """ Destroys the container. """
        container = self._get_container(create=False)
        if container is None:
            logger.info("Container doesn't exist, nothing to destroy.")
            return

        # Halts the container...
        self.halt()
        # ... and destroy it!
        logger.info('Destroying container "{name}"...'.format(name=self.name))
        container.delete(wait=True)
        logger.info('Container "{name}" destroyed!'.format(name=self.name))

    def halt(self):
        """ Stops the container. """
        if self.is_stopped:
            logger.info('The container is already stopped.')
            return

        # Removes configurations related to container's hostnames if applicable.
        self._unsetup_hostnames()

        logger.info('Stopping...')
        try:
            self._container.stop(timeout=30, force=False, wait=True)
        except LXDAPIException:
            logger.warn("Can't stop the container. Forcing...")
            self._container.stop(force=True, wait=True)

    @must_be_running
    def provision(self, barebone=None):
        """ Provisions the container. """
        if barebone is None:  # None == only if the container isn't provisioned.
            barebone = not self.is_provisioned

        if barebone:
            self._perform_barebones_setup()

        logger.info('Provisioning container "{name}"...'.format(name=self.name))

        for provisioning_item in self.options.get('provisioning', []):
            provisioning_type = provisioning_item['type'].lower()
            provisioner_class = Provisioner.provisioners.get(provisioning_type)
            if provisioner_class is not None:
                provisioner = provisioner_class(self._container, provisioning_item)
                logger.info('Provisioning with {0}'.format(provisioning_item['type']))
                provisioner.provision()

        self._container.config['user.lxdock.provisioned'] = 'true'
        self._container.save(wait=True)

    @must_be_running
    def shell(self, username=None):
        """ Opens a new interactive shell in the container. """
        # For now, it's much easier to call `lxc`, but eventually, we might want to contribute
        # to pylxd so it supports `interactive = True` in `exec()`.
        shellcfg = self.options.get('shell', {})
        shelluser = username or shellcfg.get('user')
        if shelluser:
            # This part is the result of quite a bit of `su` args trial-and-error.
            shellhome = shellcfg.get('home') if not username else None
            homearg = '--env HOME={}'.format(shellhome) if shellhome else ''
            cmd = 'lxc exec {} {} -- su -m {}'.format(self.lxd_name, homearg, shelluser)
            subprocess.call(cmd, shell=True)
        else:
            cmd = 'lxc exec {} -- su -m root'.format(self.lxd_name)
            subprocess.call(cmd, shell=True)

    def up(self):
        """ Creates, starts and provisions the container. """
        if self.is_running:
            logger.info('Container "{name}" is already running'.format(name=self.name))
            return

        logger.info('Starting container "{name}"...'.format(name=self.name))
        self._container.start(wait=True)
        if not self.is_running:
            logger.error('Something went wrong trying to start the container.')
            raise ContainerOperationFailed()

        ip = self._setup_ip()
        if not ip:
            return

        logger.info('Container "{name}" is up! IP: {ip}'.format(name=self.name, ip=ip))

        # Setup hostnames if applicable.
        self._setup_hostnames(ip)

        # Setup users if applicable.
        self._setup_users()

        # Setup shares if applicable.
        self._setup_shares()

        # Provisions the container if applicable.
        if not self.is_provisioned:
            self.provision(barebone=True)
        else:
            logger.info(
                'Container "{name}" already provisioned, not provisioning.'.format(name=self.name))

    ##################################
    # UTILITY METHODS AND PROPERTIES #
    ##################################

    @property
    def exists(self):
        """ Returns True if the considered container has already been created. """
        try:
            self.client.containers.get(self.lxd_name)
        except NotFound:
            return False
        else:
            return True

    @property
    def is_privileged(self):
        """ Returns a boolean indicating if the container is privileged. """
        return self._container.config.get('security.privileged') == 'true'

    @property
    def is_provisioned(self):
        """ Returns a boolean indicating if the container is provisioned. """
        return self._container.config.get('user.lxdock.provisioned') == 'true'

    @property
    def is_running(self):
        """ Returns a boolean indicating if the container is running. """
        return self._container.status_code == constants.CONTAINER_RUNNING

    @property
    def is_stopped(self):
        """ Returns a boolean indicating if the container is stopped. """
        return self._container.status_code == constants.CONTAINER_STOPPED

    @property
    def lxd_name(self):
        """ Returns the name of the container that is used in the scope of LXD.

        This name id supposed to be unique among all the containers managed by LXD.
        """
        # Note: all container names must be a valid hostname! That is: maximum 63 characters, no
        # dots, no digit at first position, made entirely of letters/digits/hyphens, ...
        if not hasattr(self, '_lxd_name'):
            lxd_name_prefix = '{project_name}-{name}'.format(
                project_name=self.project_name, name=self.name)
            # We compute a project ID based on inode numbers in order to ensure that our LXD names
            # are unique.
            project_id = folderid(self.homedir)
            self._lxd_name = '{prefix}-{id}'.format(
                prefix=lxd_name_prefix[:63 - len(project_id)], id=project_id)
        return self._lxd_name

    @property
    def name(self):
        """ Returns the "local" name of the container. """
        return self.options['name']

    @property
    def status(self):
        """ Returns a string identifier representing the current status of the container. """
        default_status = 'undefined'  # Note: this status should not be displayed at all...
        container = self._get_container(create=False)
        if container is None:
            status = 'not-created'
        else:
            status = {
                constants.CONTAINER_RUNNING: 'running',
                constants.CONTAINER_STOPPED: 'stopped',
            }.get(container.status_code, default_status)
        return status

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _get_container(self, create=True):
        """ Gets or creates the PyLXD container. """
        try:
            container = self.client.containers.get(self.lxd_name)
        except NotFound:
            container = None
        else:
            return container

        if not create:
            return

        logger.warn('Unable to find container "{name}" for directory "{homedir}"'.format(
            name=self.name, homedir=self.homedir))

        logger.info(
            'Creating new container "{name}" '
            'from image {image}'.format(name=self.lxd_name, image=self.options['image']))
        privileged = self.options.get('privileged', False)
        mode = self.options.get('mode', 'pull')
        container_config = {
            'name': self.lxd_name,
            'source': {
                'alias': self.options['image'],
                # The 'mode' defines how the container will be retrieved. In "local" mode the image
                # will be determined using a local alias. In "pull" mode the image will be fetched
                # from a remote server using a remote alias.
                'mode': mode,
                # The 'protocol' to use. LXD supports two protocol: 'lxd' (RESTful API that is used
                # between the clients and a LXD daemon) and 'simplestreams' (an image server
                # description format, using JSON to describe a list of images and allowing to get
                # image information and import images). We use "simplestreams" by default (as the
                # lxc command do).
                'protocol': self.options.get('protocol', 'simplestreams'),
                # The 'server' that should be used to fetch the images. We use the default
                # linuxcontainers server for LXC and LXD when no value is provided (and if we are
                # not in "local" mode).
                'server': (self.options.get('server', self._default_image_server) if mode == 'pull'
                           else ''),
                'type': 'image',
            },
            'config': {
                'security.privileged': 'true' if privileged else 'false',
                'user.lxdock.made': '1',
                'user.lxdock.homedir': self.homedir,
            },
        }
        try:
            return self.client.containers.create(container_config, wait=True)
        except LXDAPIException as e:
            logger.error("Can't create container: {error}".format(error=e))
            raise ContainerOperationFailed()

    def _perform_barebones_setup(self):
        """ Performs bare bones setup on the machine. """
        logger.info('Doing bare bones setup on the machine...')

        # First install bare bones packages on the container.
        self._guest.install_barebones_packages()

        # Then add the current user's SSH pubkey to the container's root SSH config.
        ssh_pubkey = self._host.get_ssh_pubkey()
        if ssh_pubkey is not None:
            self._guest.add_ssh_pubkey_to_root_authorized_keys(ssh_pubkey)
        else:
            logger.warning('SSH pubkey was not found. Provisioning tools may not work correctly...')

    def _setup_hostnames(self, ip):
        """ Configure the potential hostnames associated with the container. """
        hostnames = self.options.get('hostnames', [])
        if not hostnames:
            return

        etchosts = EtcHosts()
        for hostname in hostnames:
            logger.info('Setting {hostname} to point to {ip}.'.format(
                hostname=hostname, ip=ip))
            etchosts.ensure_binding_present(hostname, ip)
        if etchosts.changed:
            logger.info("Saving host bindings to /etc/hosts. sudo may be needed")
            etchosts.save()

    def _setup_ip(self):
        """ Setup the IP address of the considered container. """
        ip = get_ipv4_ip(self._container)
        if not ip:
            logger.info('No IP yet, waiting for at most 10 seconds...')
            ip = self._wait_for_ipv4_ip()
        if not ip:
            logger.warn('STILL no IP! Container is up, but probably broken.')
            logger.info('Maybe that restarting it will help? Not trying to provision.')
        return ip

    def _setup_shares(self):
        """ Setup the shared folders associated with the container. """
        if 'shares' not in self.options:
            return

        logger.info('Setting up shares...')

        container = self._container

        # First, let's make an inventory of shared sources that were already there.
        existing_shares = {
            k: d for k, d in container.devices.items() if k.startswith('lxdockshare')}
        existing_sources = {d['source'] for d in existing_shares.values()}

        # Let's get rid of previously set up lxdock shares.
        for k in existing_shares:
            del container.devices[k]

        for i, share in enumerate(self.options.get('shares', []), start=1):
            source = os.path.join(self.homedir, share['source'])
            if source not in existing_sources:
                logger.info('Setting host-side ACL for {}'.format(source))
                self._host.give_current_user_access_to_share(source)
                if not self.is_privileged:
                    # We are considering a safe container. So give the mapped root user permissions
                    # to read/write contents in the shared folders too.
                    self._host.give_mapped_user_access_to_share(source)
                    # We also give these permissions to any user that was created with LXDock.
                    for uconfig in self.options.get('users', []):
                        username = uconfig.get('name')
                        self._host.give_mapped_user_access_to_share(
                            source, userpath=uconfig.get('home', '/home/' + username))

            shareconf = {'type': 'disk', 'source': source, 'path': share['dest'], }
            container.devices['lxdockshare%s' % i] = shareconf
        container.save(wait=True)

    def _setup_users(self):
        """ Creates users defined in the container's options if applicable. """
        users = self.options.get('users', [])
        if not users:
            return

        logger.info('Ensuring users are created...')
        for user_config in users:
            name = user_config.get('name')
            home = user_config.get('home')
            self._guest.create_user(name, home=home)

    def _unsetup_hostnames(self):
        """ Removes the configuration associated with the hostnames of the container. """
        hostnames = self.options.get('hostnames', [])
        if not hostnames:
            return

        etchosts = EtcHosts()
        for hostname in hostnames:
            logger.info('Unsetting {hostname}. sudo needed.'.format(hostname=hostname))
            etchosts.ensure_binding_absent(hostname)
        if etchosts.changed:
            etchosts.save()

    def _wait_for_ipv4_ip(self, seconds=10):
        """ Waits some time before trying to get the IP of the container and returning it. """
        for i in range(seconds):
            time.sleep(1)
            ip = get_ipv4_ip(self._container)
            if ip:
                return ip
        return ''

    @property
    def _container(self):
        """ Returns the PyLXD Container instance associated with the considered container. """
        if not hasattr(self, '_pylxd_container'):
            self._pylxd_container = self._get_container()
        return self._pylxd_container

    @property
    def _guest(self):
        """ Returns the `Guest` instance associated with the considered container.

        None is returned if no guest class can be determined for the considered container. """
        if not hasattr(self, '_container_guest'):
            guest_class = next((k for k in Guest.guests if k.detect(self._container)), Guest)
            self._container_guest = guest_class(self._container)
        return self._container_guest

    @property
    def _host(self):
        """ Returns the `Host` instance associated with the considered host.

        None is returned if no guest class can be determined for the considered container. """
        if not hasattr(self, '_container_host'):
            host_class = next((k for k in Host.hosts if k.detect()), Host)
            self._container_host = host_class(self._container)
        return self._container_host



import unittest.mock

from lxdock.provisioners import AnsibleProvisioner


class TestAnsibleProvisioner:
    @unittest.mock.patch('subprocess.Popen')
    @unittest.mock.patch('tempfile.NamedTemporaryFile')
    def test_can_run_ansible_playbooks(self, mock_tempfile, mock_popen):
        class MockedTMPFile:
            name = 'tmpfile'

            def flush(self):
                pass

            def write(self, data):
                pass

        mock_tempfile.return_value.__enter__.return_value = MockedTMPFile()
        lxd_container = unittest.mock.Mock()
        lxd_state = unittest.mock.Mock()
        lxd_state.network.__getitem__ = unittest.mock.MagicMock(
            return_value={'addresses': [{'family': 'init', 'address': '0.0.0.0', }, ]})
        lxd_container.state.return_value = lxd_state
        provisioner = AnsibleProvisioner(lxd_container, {'playbook': 'deploy.yml'})
        provisioner.provision()
        assert mock_popen.call_args[0] == (
            'ANSIBLE_HOST_KEY_CHECKING=False ansible-playbook -i tmpfile deploy.yml', )



import logging
import sys

from colorlog import ColoredFormatter

LOG_COLORS = {
    'DEBUG':    'reset',
    'INFO':     'reset',
    'WARNING':  'yellow',
    'ERROR':    'bold_red,bg_white',
    'CRITICAL': 'bold_red,bg_white',
}


class _AtMostWarningFilter(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.ERROR


class _AtleastErrorFilter(logging.Filter):
    def filter(self, record):
        return record.levelno >= logging.ERROR


def get_default_formatter():
    """ Returns the default formatter used to log messages for LXDock. """
    return ColoredFormatter('%(log_color)s%(message)s', log_colors=LOG_COLORS)


def get_per_container_formatter(container_name):
    """ Returns a logging formatter which prefixes each message with a container name. """
    return ColoredFormatter(
        '%(log_color)s==> {name}: %(message)s'.format(name=container_name), log_colors=LOG_COLORS)


logger = logging.getLogger(__name__)

console_stdout_handler = logging.StreamHandler(sys.stdout)
console_stdout_handler.addFilter(_AtMostWarningFilter())
console_stderr_handler = logging.StreamHandler(sys.stderr)
console_stderr_handler.addFilter(_AtleastErrorFilter())

console_stdout_handler.setFormatter(get_default_formatter())
console_stderr_handler.setFormatter(get_default_formatter())



from voluptuous import All, In, IsDir, Length, Required, Schema, Url

from .validators import Hostname, LXDIdentifier


_top_level_and_containers_common_options = {
    'hostnames': [Hostname(), ],
    'image': str,
    'mode': In(['local', 'pull', ]),
    'privileged': bool,
    'protocol': In(['lxd', 'simplestreams', ]),
    'provisioning': [{
        # Common options
        'type': str,

        # Ansible specific options
        'playbook': str,
    }],
    'server': Url(),
    'shares': [{
        # The existence of the source directory will be checked!
        'source': IsDir(),
        'dest': str,
    }],
    'shell': {
        'user': str,
        'home': str,
    },
    'users': [{
        # Usernames max length is set 32 characters according to useradd's man page.
        Required('name'): All(str, Length(max=32)),
        'home': str,
    }],
}

_container_options = {
    Required('name'): LXDIdentifier(),
}
_container_options.update(_top_level_and_containers_common_options)

_lxdock_options = {
    Required('name'): LXDIdentifier(),
    'containers': [_container_options, ],
}
_lxdock_options.update(_top_level_and_containers_common_options)

# The schema will be used to validate LXDock files!
schema = Schema(_lxdock_options)



from .base import Guest


class AlpineGuest(Guest):
    """ This guest can provision Alpine containers. """

    name = 'alpine'
    barebones_packages = [
        'openssh',
        'python',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['apk', 'update'])
        self.run(['apk', 'add'] + self.barebones_packages)
        self.run(['rc-update', 'add', 'sshd'])
        self.run(['/etc/init.d/sshd', 'start'])



import os
import unittest.mock

import pytest

from lxdock.cli.constants import INIT_LXDOCK_FILE_CONTENT
from lxdock.cli.main import LXDock, main
from lxdock.cli.project import get_project
from lxdock.conf.config import Config
from lxdock.conf.exceptions import ConfigError
from lxdock.container import Container
from lxdock.exceptions import LXDockException
from lxdock.project import Project


FIXTURE_ROOT = os.path.join(os.path.dirname(__file__), 'fixtures')


@unittest.mock.patch.object(LXDock, 'help')
def test_main_function_can_run_a_lxdock_command(mock_help_action):
    main(['help'])
    assert mock_help_action.call_count == 1


class TestLXDock:
    @unittest.mock.patch.object(LXDock, 'project_config')
    @unittest.mock.patch.object(Config, 'serialize')
    def test_can_display_the_config_file_of_the_project(self, serialize_mock, mock_config):
        mock_config.__get__ = unittest.mock.Mock(
            return_value=Config.from_base_dir(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['config'])
        assert serialize_mock.call_count == 1

    @unittest.mock.patch.object(LXDock, 'project_config')
    @unittest.mock.patch.object(Config, 'serialize')
    def test_can_display_the_containers_of_the_config_file_of_the_project(
            self, serialize_mock, mock_config):
        mock_config.__get__ = unittest.mock.Mock(
            return_value=Config.from_base_dir(os.path.join(FIXTURE_ROOT, 'project01')))
        argv = ['config', '--containers']
        n = LXDock(argv)
        assert serialize_mock.call_count == 0
        assert n._parsers['main'].parse_args(argv).containers

    @unittest.mock.patch('builtins.input', return_value='Y')
    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'destroy')
    @unittest.mock.patch.object(Container, 'exists')
    def test_can_run_the_destroy_action_for_all_containers_of_a_project(
            self, container_exists_mock, mock_project_destroy, mock_project, y_mock):
        container_exists_mock.__get__ = unittest.mock.Mock(return_value=True)
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['destroy'])
        assert mock_project_destroy.call_count == 1
        assert mock_project_destroy.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch('builtins.input', return_value='Y')
    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'destroy')
    @unittest.mock.patch.object(Container, 'exists')
    def test_can_run_the_destroy_action_for_specific_containers(
            self, container_exists_mock, mock_project_destroy, mock_project, y_mock):
        container_exists_mock.__get__ = unittest.mock.Mock(return_value=True)
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['destroy', 'default'])
        assert mock_project_destroy.call_count == 1
        assert mock_project_destroy.call_args == [{'container_names': ['default', ], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'destroy')
    @unittest.mock.patch.object(Container, 'exists')
    def test_can_run_the_destroy_containers_using_a_force_option(
            self, container_exists_mock, mock_project_destroy, mock_project):
        container_exists_mock.__get__ = unittest.mock.Mock(return_value=True)
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['destroy', '--force'])
        assert mock_project_destroy.call_count == 1
        assert mock_project_destroy.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'destroy')
    @unittest.mock.patch.object(Container, 'exists')
    def test_can_force_the_destroy_action_if_the_containers_do_not_exist(
            self, container_exists_mock, mock_project_destroy, mock_project):
        container_exists_mock.__get__ = unittest.mock.Mock(return_value=False)
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['destroy'])
        assert mock_project_destroy.call_count == 1
        assert mock_project_destroy.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'halt')
    def test_can_run_the_halt_action_for_all_containers_of_a_project(
            self, mock_project_halt, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['halt'])
        assert mock_project_halt.call_count == 1
        assert mock_project_halt.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'halt')
    def test_can_run_the_halt_action_for_specific_containers(
            self, mock_project_halt, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['halt', 'c1', 'c2'])
        assert mock_project_halt.call_count == 1
        assert mock_project_halt.call_args == [{'container_names': ['c1', 'c2', ], }, ]

    def test_can_print_the_global_help(self):
        argv = ['help']
        n = LXDock(argv)
        assert n._parsers['main'].parse_args(argv).subcommand is None

    def test_can_print_the_help_for_a_specific_subcommand(self):
        argv = ['help', 'up']
        n = LXDock(argv)
        assert n._parsers['main'].parse_args(argv).subcommand == 'up'

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'provision')
    def test_can_run_the_provision_action_for_all_containers_of_a_project(
            self, mock_project_provision, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['provision'])
        assert mock_project_provision.call_count == 1
        assert mock_project_provision.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'provision')
    def test_can_run_the_provision_action_for_specific_containers(
            self, mock_project_provision, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['provision', 'c1', 'c2'])
        assert mock_project_provision.call_count == 1
        assert mock_project_provision.call_args == [{'container_names': ['c1', 'c2', ], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'shell')
    def test_can_run_the_shell_action_for_all_containers_of_a_project(
            self, mock_project_shell, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['shell'])
        assert mock_project_shell.call_count == 1
        assert mock_project_shell.call_args == [{'container_name': None, 'username': None}, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'shell')
    def test_can_run_the_shell_action_for_a_specific_container(
            self, mock_project_shell, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['shell', 'c1'])
        assert mock_project_shell.call_count == 1
        assert mock_project_shell.call_args == [{'container_name': 'c1', 'username': None}, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'shell')
    def test_can_run_the_shell_action_for_a_specific_user(self, mock_project_shell, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['shell', '-u', 'foobar'])
        assert mock_project_shell.call_count == 1
        assert mock_project_shell.call_args == [{'container_name': None, 'username': 'foobar'}, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'status')
    def test_can_run_the_status_action_for_all_containers_of_a_project(
            self, mock_project_status, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['status'])
        assert mock_project_status.call_count == 1
        assert mock_project_status.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'status')
    def test_can_run_the_status_action_for_specific_containers(
            self, mock_project_status, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['status', 'c1', 'c2'])
        assert mock_project_status.call_count == 1
        assert mock_project_status.call_args == [{'container_names': ['c1', 'c2', ], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'up')
    def test_can_run_the_up_action_for_all_containers_of_a_project(
            self, mock_project_up, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['up'])
        assert mock_project_up.call_count == 1
        assert mock_project_up.call_args == [{'container_names': [], }, ]

    @unittest.mock.patch.object(LXDock, 'project')
    @unittest.mock.patch.object(Project, 'up')
    def test_can_run_the_up_action_for_specific_containers(
            self, mock_project_up, mock_project):
        mock_project.__get__ = unittest.mock.Mock(
            return_value=get_project(os.path.join(FIXTURE_ROOT, 'project01')))
        LXDock(['up', 'c1', 'c2'])
        assert mock_project_up.call_count == 1
        assert mock_project_up.call_args == [{'container_names': ['c1', 'c2', ], }, ]

    def test_exit_if_no_action_is_provided(self):
        with pytest.raises(SystemExit):
            LXDock([])

    def test_exit_if_a_cli_error_is_encountered(self):
        with pytest.raises(SystemExit):
            LXDock(['help', 'unknown'])

    @unittest.mock.patch.object(LXDock, 'project')
    def test_exit_if_a_keyboard_interrupt_occurs(self, mock_project):
        mock_project.__get__ = unittest.mock.Mock(side_effect=KeyboardInterrupt)
        with pytest.raises(SystemExit):
            LXDock(['destroy'])

    @unittest.mock.patch.object(LXDock, 'project')
    def test_exit_if_a_config_error_is_encountered(self, mock_project):
        mock_project.__get__ = unittest.mock.Mock(side_effect=ConfigError(msg='TEST'))
        with pytest.raises(SystemExit):
            LXDock(['up'])

    @unittest.mock.patch.object(LXDock, 'project')
    def test_exit_if_a_lxdock_error_is_encountered(self, mock_project):
        mock_project.__get__ = unittest.mock.Mock(side_effect=LXDockException(msg='TEST'))
        with pytest.raises(SystemExit):
            LXDock(['up'])

    @unittest.mock.patch.object(Config, 'from_base_dir')
    def test_project_config_property_works(
            self, from_base_dir_mock):
        n = LXDock(['config'])
        config1, config2 = n.project_config, n.project_config
        assert config1 == config2
        assert from_base_dir_mock.call_count == 1

    @unittest.mock.patch.object(Config, 'from_base_dir')
    @unittest.mock.patch.object(Project, 'from_config')
    def test_project_property_works(
            self, from_config_mock, from_base_dir_mock):
        n = LXDock(['config'])
        project1, project2 = n.project, n.project
        assert project1 == project2
        assert from_config_mock.call_count == 1

    @unittest.mock.patch('builtins.open')
    def test_can_generate_a_basic_lxdock_file(self, mock_open):
        fd_mock = unittest.mock.Mock()
        mock_open.return_value.__enter__.return_value = fd_mock
        LXDock(['init', ])
        assert mock_open.call_count == 1
        assert mock_open.call_args[0] == ('lxdock.yml', )
        assert fd_mock.write.call_count == 1
        assert fd_mock.write.call_args[0][0] == INIT_LXDOCK_FILE_CONTENT.format(
            project_name=os.path.split(os.getcwd())[1], image='ubuntu/xenial')

    @unittest.mock.patch('builtins.open')
    @unittest.mock.patch('os.getcwd')
    def test_cannot_generate_a_lxdock_file_if_there_is_already_an_existing_lxdock_file(
            self, mock_getcwd, mock_open):
        mock_getcwd.return_value = os.path.join(FIXTURE_ROOT, 'project01')
        fd_mock = unittest.mock.Mock()
        mock_open.return_value.__enter__.return_value = fd_mock
        with pytest.raises(SystemExit):
            LXDock(['init', ])

    @unittest.mock.patch('builtins.open')
    @unittest.mock.patch('os.getcwd')
    def test_can_generate_a_lxdock_file_by_overwritting_an_existing_file_with_the_force_option(
            self, mock_getcwd, mock_open):
        mock_getcwd.return_value = os.path.join(FIXTURE_ROOT, 'project01')
        fd_mock = unittest.mock.Mock()
        mock_open.return_value.__enter__.return_value = fd_mock
        LXDock(['init', '--force'])
        assert mock_open.call_count == 1
        assert mock_open.call_args[0] == ('lxdock.yml', )
        assert fd_mock.write.call_count == 1
        assert fd_mock.write.call_args[0][0] == INIT_LXDOCK_FILE_CONTENT.format(
            project_name=os.path.split(os.getcwd())[1], image='ubuntu/xenial')

    @unittest.mock.patch('builtins.open')
    def test_can_generate_a_lxdock_file_with_a_custom_image(self, mock_open):
        fd_mock = unittest.mock.Mock()
        mock_open.return_value.__enter__.return_value = fd_mock
        LXDock(['init', '--image', 'debian/jessie', ])
        assert mock_open.call_count == 1
        assert mock_open.call_args[0] == ('lxdock.yml', )
        assert fd_mock.write.call_count == 1
        assert fd_mock.write.call_args[0][0] == INIT_LXDOCK_FILE_CONTENT.format(
            project_name=os.path.split(os.getcwd())[1], image='debian/jessie')

    @unittest.mock.patch('builtins.open')
    def test_can_generate_a_lxdock_file_with_a_custom_project_name(self, mock_open):
        fd_mock = unittest.mock.Mock()
        mock_open.return_value.__enter__.return_value = fd_mock
        LXDock(['init', '--project', 'customproject', ])
        assert mock_open.call_count == 1
        assert mock_open.call_args[0] == ('lxdock.yml', )
        assert fd_mock.write.call_count == 1
        assert fd_mock.write.call_args[0][0] == INIT_LXDOCK_FILE_CONTENT.format(
            project_name='customproject', image='ubuntu/xenial')



import unittest.mock

from lxdock.guests import DebianGuest


class TestDebianGuest:
    def test_can_install_barebones_packages(self):
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = DebianGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == \
            (['apt-get', 'install', '-y'] + DebianGuest.barebones_packages, )



class LXDockException(Exception):
    def __init__(self, msg=None):
        self.msg = msg


class ProjectError(LXDockException):
    """ An error occured while excuting some action at the project level. """


class ContainerOperationFailed(LXDockException):
    """ An operation on a specific container failed. """



"""
    Base provisioner
    ================
    This module provides the `Provisioner` base class that is used to define the provisioners
    supported by LXDock (eg. Ansible, ...).
"""

from ..utils.metaclass import with_metaclass

__all__ = ['Provisioner', ]


class InvalidProvisioner(Exception):
    """ The `Provisioner` subclass is not valid. """


class _ProvisionerBase(type):
    """ Metaclass for all LXDock provisioners.

    This metaclass ensures that all defined `Provisioner` subclasses have the required attributes
    and proceeds to some validation checks. Additionally it implements the "plugin mount" paradigm
    and stores a dictionary of `Provisioner` subclasses in the namespace of the "plugin mount"
    class (`Provisioner`).
    """

    def __new__(cls, name, bases, attrs):
        super_new = super(_ProvisionerBase, cls).__new__
        parents = [base for base in bases if isinstance(base, _ProvisionerBase)]

        # We stop here if we are considering the top-level class to which this metaclass was applied
        # and not one of its subclasses (eg. Provisioner).
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Constructs the Provisioner class.
        new_provisioner = super_new(cls, name, bases, attrs)

        # Performs some validation checks.
        if not new_provisioner.name:
            raise InvalidProvisioner(
                "The 'name' attribute of Provisioner subclasses cannot be None")

        return new_provisioner

    def __init__(cls, name, bases, attrs):
        # We implement the "mount point" paradigm here in order to make all `Provisioner`
        # subclassses available in a single attribute.
        if not hasattr(cls, 'provisioners'):
            # The class has no provisioners dict: this means that we are considering the
            # "plugin mount" class. So we created the dict that will hold all the defined
            # `Provisioner` subclasses.
            cls.provisioners = {}
        else:
            # The `provisioners` attribute already exists so we are considering a `Provisioner`
            # subclass, which needs to be registered.
            cls.provisioners.update({cls.name.lower(): cls})


class Provisioner(with_metaclass(_ProvisionerBase)):
    """ Represents a single provisioner.

    `Provisioner` subclasses will be used by `Container` instances to run provisioning operations
    associated with the considered containers. For example they can be used to run Ansible playbook
    to provision a web application on the container.
    """

    # The `name` of a provisioner is a required attribute and should always be set on `Provisioner`
    # subclasses.
    name = None

    def __init__(self, lxd_container, options):
        self.lxd_container = lxd_container
        self.options = options.copy()

    def provision(self):
        """ Performs the provisioning operations using the considered provisioner. """
        # This method should be overriden in `Provisioner` subclasses.



import unittest.mock

from lxdock.guests import OpenSUSEGuest


class TestOpenSUSEGuest:
    def test_can_install_barebones_packages(self):
        lxd_container = unittest.mock.Mock()
        lxd_container.execute.return_value = ('ok', 'ok', '')
        guest = OpenSUSEGuest(lxd_container)
        guest.install_barebones_packages()
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args[0] == \
            (['zypper', '--non-interactive', 'install'] +
             OpenSUSEGuest.barebones_packages, )



from .base import Guest


class FedoraGuest(Guest):
    """ This guest can provision Fedora containers. """

    name = 'fedora'
    barebones_packages = [
        'openssh-server',
        'python3',
    ]

    def install_barebones_packages(self):
        """ Installs packages when the guest is first provisionned. """
        self.run(['dnf', '-y', 'install', ] + self.barebones_packages)



import os
import pytest

from lxdock.conf.config import Config
from lxdock.conf.exceptions import ConfigFileNotFoundError, ConfigFileValidationError


FIXTURE_ROOT = os.path.join(os.path.dirname(__file__), 'fixtures')


class TestConfig:
    def test_can_be_initialized_from_a_base_dir(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config['name'] == 'project01'
        assert config['image'] == 'ubuntu/xenial'
        assert config['mode'] == 'pull'

    def test_works_if_multiple_config_files_are_found(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_multiple_config_files')
        config = Config.from_base_dir(project_dir)
        assert config.filename == os.path.join(project_dir, 'lxdock.yml')

    def test_raises_an_error_if_the_config_file_cannot_be_found(self):
        project_dir = os.path.dirname(__file__)
        with pytest.raises(ConfigFileNotFoundError):
            Config.from_base_dir(project_dir)

    def test_raises_an_error_if_the_config_file_is_invalid(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_invalid_config')
        with pytest.raises(ConfigFileValidationError):
            Config.from_base_dir(project_dir)

    def test_allows_to_get_items_from_config_dict(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config['name'] == 'project01'

    def test_allows_to_check_item_presence_in_config_dict(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert 'name' in config

    def test_can_load_container_config_from_top_level_options_if_no_containers_are_defined(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config.containers == [
            {'image': 'ubuntu/xenial', 'mode': 'pull', 'name': 'default'},
        ]

    def test_can_properly_handle_configurations_defining_multiple_containers(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(project_dir)
        assert config.containers == [
            {'mode': 'pull', 'image': 'ubuntu/xenial', 'name': 'web'},
            {
                'mode': 'pull', 'hostnames': ['ci.local'], 'image': 'debian/jessie', 'name': 'ci',
                'privileged': True,
            },
        ]

    def test_can_serialize_the_parsed_config(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config.serialize() == 'image: ubuntu/xenial\nmode: pull\nname: project01\n'



"""
    Base utilities related to metaclasses
    =====================================
    This module provides utilities allowing to manage metaclasses in the context of LXDock.
    LXDock uses metaclasses to implement plugin-like components like guests, hosts, ...
"""


def with_metaclass(meta, *bases):
    """ Creates a base class with a metaclass.

    This function defines a `metaclass` class that inherits from the passed `meta` metaclass and
    returns a new class whose properties will differ depending on the class being constructed:

    * if we are considering the base class to which the `with_metaclass` function was applied, a new
      class without base classes will be constructed

    * if we are considering a subclass of the class to which the `with_metaclass` function was
      applied, the class being constructed will inherit from the `meta` metaclass

    This trick is usefull to differentiate the class to which the `with_metaclass` function was
    applied from its subclasses during the execution of the `__new__` magic method (which is defined
    inside the `meta` metaclass). This allows fine control over instances created with the `meta`
    metaclass.
    """
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('NewBase', None, {})



import logging
import subprocess
import tempfile

from ..network import get_ipv4_ip

from .base import Provisioner

logger = logging.getLogger(__name__)


class AnsibleProvisioner(Provisioner):
    """ Allows to perform provisioning operations using Ansible. """

    name = 'ansible'

    def provision(self):
        """ Performs the provisioning operations using the considered provisioner. """
        ip = get_ipv4_ip(self.lxd_container)
        with tempfile.NamedTemporaryFile() as tmpinv:
            tmpinv.write('{} ansible_user=root'.format(ip).encode('ascii'))
            tmpinv.flush()
            cmd = 'ANSIBLE_HOST_KEY_CHECKING=False ansible-playbook -i {} {}'.format(
                tmpinv.name, self.options['playbook'])
            logger.debug(cmd)
            p = subprocess.Popen(cmd, shell=True)
            p.wait()



import sys

from setuptools import find_packages
from setuptools import setup

import lxdock

if sys.version_info < (3, 4):
    sys.exit('lxdock requires Python 3.4+ to run')

setup(
    name='lxdock',
    version=lxdock.__version__,
    author='Virgil Dupras, Morgan Aubert',
    author_email='',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/lxdock/lxdock',
    license='GPLv3',
    description='LXDock is a tool for orchestrating LXD containers.',
    keywords='lxd lxc containers environments orchestration devops',
    zip_safe=False,
    install_requires=[
        'colorlog>=2.0,<3.0',
        'pylxd>=2.2,<3.0',
        'PyYAML>=3.0,<4.0',
        'voluptuous>=0.9,<1.0',
    ],
    tests_require=[
        'pytest',
    ],
    entry_points={
        'console_scripts': [
            'lxdock = lxdock.cli.main:main',
        ],
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



"""
    Provisioner components
    ======================
    This package contains provisioner definitions. Provisioners are used to perform to provision the
    containers using provisioning tools (eg. Ansible).
"""

from .ansible import *  # noqa
from .base import *  # noqa



import os
import unittest.mock

import pytest

from lxdock.conf.config import Config
from lxdock.container import Container
from lxdock.exceptions import ProjectError
from lxdock.project import Project
from lxdock.test import LXDTestCase


THIS_DIR = os.path.join(os.path.dirname(__file__))
FIXTURE_ROOT = os.path.join(os.path.dirname(__file__), 'fixtures')


class TestProject(LXDTestCase):
    def test_can_be_initialized_from_a_config_file(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project01', homedir, self.client, config)
        assert project.name == 'project01'
        assert project.homedir == homedir
        assert len(project.containers) == 1
        assert project.containers[0].name == 'default'
        assert project.containers[0].homedir == homedir
        assert project.containers[0].options['mode'] == 'pull'
        assert project.containers[0].options['image'] == 'ubuntu/xenial'

    def test_can_return_a_container_by_name(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        container = project.get_container_by_name('lxdock-pytest-web')
        assert container and container.name == 'lxdock-pytest-web'

    def test_raises_an_error_if_it_cannot_find_a_container(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        with pytest.raises(ProjectError):
            project.get_container_by_name('dummy')

    def test_can_start_all_the_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        project.up()
        for container in project.containers:
            assert container.is_running

    def test_can_start_some_specific_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        project.up(container_names=['lxdock-pytest-web'])
        container_web = project.get_container_by_name('lxdock-pytest-web')
        container_ci = project.get_container_by_name('lxdock-pytest-ci')
        assert container_web.is_running
        assert container_ci.is_stopped

    def test_can_halt_all_the_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        project.up()
        project.halt()
        for container in project.containers:
            assert container.is_stopped

    def test_can_halt_some_specific_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        project.up()
        project.halt(container_names=['lxdock-pytest-web'])
        container_web = project.get_container_by_name('lxdock-pytest-web')
        container_ci = project.get_container_by_name('lxdock-pytest-ci')
        assert container_web.is_stopped
        assert container_ci.is_running

    def test_can_provision_all_the_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        container_options = {
            'name': self.containername('dummytest'), 'image': 'ubuntu/xenial', 'mode': 'pull',
            'provisioning': [
                {'type': 'ansible',
                 'playbook': os.path.join(THIS_DIR, 'fixtures/provision_with_ansible.yml'), }
            ],
        }
        project = Project(
            'project02', homedir, self.client,
            [Container('myproject', THIS_DIR, self.client, **container_options)])
        project.up()
        project.provision()
        for container in project.containers:
            assert container.is_provisioned

    def test_can_provision_some_specific_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        container_options = {
            'name': self.containername('thisisatest'), 'image': 'ubuntu/xenial', 'mode': 'pull',
            'provisioning': [
                {'type': 'ansible',
                 'playbook': os.path.join(THIS_DIR, 'fixtures/provision_with_ansible.yml'), }
            ],
        }
        project = Project(
            'project02', homedir, self.client,
            [Container('myproject', THIS_DIR, self.client, **container_options)])
        project.up()
        project.provision(container_names=['lxdock-pytest-thisisatest'])
        container_web = project.get_container_by_name('lxdock-pytest-thisisatest')
        assert container_web.is_provisioned

    def test_can_destroy_all_the_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        container_options = {
            'name': self.containername('dummytest'), 'image': 'ubuntu/xenial', 'mode': 'pull',
            'provisioning': [
                {'type': 'ansible',
                 'playbook': os.path.join(THIS_DIR, 'fixtures/provision_with_ansible.yml'), }
            ],
        }
        project = Project(
            'project02', homedir, self.client,
            [Container('myproject', THIS_DIR, self.client, **container_options)])
        project.up()
        project.destroy()
        for container in project.containers:
            assert not container.exists

    def test_can_destroy_some_specific_containers_of_a_project(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        project.up()
        project.destroy(container_names=['lxdock-pytest-web'])
        container_web = project.get_container_by_name('lxdock-pytest-web')
        container_ci = project.get_container_by_name('lxdock-pytest-ci')
        assert not container_web.exists
        assert container_ci.exists

    def test_cannot_open_shell_into_many_containers(self):
        homedir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(homedir)
        project = Project.from_config('project02', homedir, self.client, config)
        with pytest.raises(ProjectError):
            project.shell()

    @unittest.mock.patch('subprocess.call')
    def test_can_open_a_shell_for_a_specific_container(self, mocked_call, persistent_container):
        homedir = os.path.join(FIXTURE_ROOT, 'project03')
        project = Project('project02', homedir, self.client, [persistent_container, ])
        project.shell(container_name='testcase-persistent')
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} -- su -m root'.format(persistent_container.lxd_name)
